# Copyright (c) 2019 AT&T Intellectual Property.
# Copyright (c) 2018-2019 Nokia.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# This source code is part of the near-RT RIC (RAN Intelligent Controller)
# platform project (RICP).
#


"""The module provides implementation of Shared Data Layer (SDL) configurability."""
import os
from enum import Enum
from collections import namedtuple


class DbBackendType(Enum):
    """Enumeration class of supported SDL database backend types."""
    REDIS = 1
    FAKE_DICT = 2


class _Configuration():
    """This class implements Shared Data Layer (SDL) configurability."""
    Params = namedtuple('Params', ['db_host', 'db_port', 'db_sentinel_port',
                                   'db_sentinel_master_name', 'db_type'])

    def __init__(self, fake_db_backend):
        self.params = self._read_configuration(fake_db_backend)

    def __str__(self):
        return str(
            {
                "DB host": self.params.db_host,
                "DB port": self.params.db_port,
                "DB master sentinel": self.params.db_sentinel_master_name,
                "DB sentinel port": self.params.db_sentinel_port,
                "DB type": self.params.db_type.name,
            }
        )

    def get_params(self):
        """Return SDL configuration."""
        return self.params

    @classmethod
    def _read_configuration(cls, fake_db_backend):
        backend_type = DbBackendType.REDIS
        if fake_db_backend:
            if fake_db_backend.lower() != 'dict':
                msg = ("Configuration error: "
                       "SDL instance was initiated with wrong "
                       "'fake_db_backend' argument value: {}. "
                       "Value 'dict' is only supported.".
                       format(fake_db_backend))
                raise ValueError(msg)

            backend_type = DbBackendType.FAKE_DICT

        return _Configuration.Params(db_host=os.getenv('DBAAS_SERVICE_HOST'),
                                     db_port=os.getenv('DBAAS_SERVICE_PORT'),
                                     db_sentinel_port=os.getenv('DBAAS_SERVICE_SENTINEL_PORT'),
                                     db_sentinel_master_name=os.getenv('DBAAS_MASTER_NAME'),
                                     db_type=backend_type)



# Copyright (c) 2019 AT&T Intellectual Property.
# Copyright (c) 2018-2019 Nokia.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# This source code is part of the near-RT RIC (RAN Intelligent Controller)
# platform project (RICP).
#


"""The module provides fake implementation of Shared Data Layer (SDL) database backend interface."""
import fnmatch
from typing import (Callable, Dict, Set, List, Optional, Tuple, Union)
import queue
import threading
import time
from ricsdl.configuration import _Configuration
from .dbbackend_abc import DbBackendAbc
from .dbbackend_abc import DbBackendLockAbc


class FakeDictBackend(DbBackendAbc):
    """
    A class providing fake implementation of database backend of Shared Data Layer (SDL).
    This class does not provide working database solution, this class can be used in testing
    purposes only. Implementation does not provide shared database resource, SDL client sees
    only its local local 'fake' database, which is a simple Python dictionary. Also keys are
    stored in database under the same namespace.

    Args:
        configuration (_Configuration): SDL configuration, containing credentials to connect to
                                        Redis database backend.
    """
    def __init__(self, configuration: _Configuration) -> None:
        super().__init__()
        self._db = {}
        self._configuration = configuration
        self._queue = queue.Queue(1)
        self._channel_cbs = {}
        self._listen_thread = threading.Thread(target=self._listen, daemon=True)
        self._run_in_thread = False

    def __str__(self):
        return str(
            {
                "DB type": "FAKE DB",
            }
        )

    def is_connected(self):
        return True

    def close(self):
        pass

    def set(self, ns: str, data_map: Dict[str, bytes]) -> None:
        self._db.update(data_map.copy())

    def set_if(self, ns: str, key: str, old_data: bytes, new_data: bytes) -> bool:
        if key not in self._db:
            return False
        db_data = self._db[key]
        if db_data == old_data:
            self._db[key] = new_data
            return True
        return False

    def set_if_not_exists(self, ns: str, key: str, data: bytes) -> bool:
        if key not in self._db:
            self._db[key] = data
            return True
        return False

    def get(self, ns: str, keys: List[str]) -> Dict[str, bytes]:
        ret = {}
        for k in keys:
            if k in self._db:
                ret[k] = self._db[k]
        return ret

    def find_keys(self, ns: str, key_pattern: str) -> List[str]:
        ret = []
        for k in self._db:
            if fnmatch.fnmatch(k, key_pattern):
                ret.append(k)
        return ret

    def find_and_get(self, ns: str, key_pattern: str) -> Dict[str, bytes]:
        ret = {}
        for key, val in self._db.items():
            if fnmatch.fnmatch(key, key_pattern):
                ret[key] = val
        return ret

    def remove(self, ns: str, keys: List[str]) -> None:
        for key in keys:
            self._db.pop(key, None)

    def remove_if(self, ns: str, key: str, data: bytes) -> bool:
        if key in self._db:
            db_data = self._db[key]
            if db_data == data:
                self._db.pop(key)
                return True
        return False

    def add_member(self, ns: str, group: str, members: Set[bytes]) -> None:
        if group in self._db:
            self._db[group] = self._db[group] | members.copy()
        else:
            self._db[group] = members.copy()

    def remove_member(self, ns: str, group: str, members: Set[bytes]) -> None:
        if group not in self._db:
            return
        for member in members:
            self._db[group].discard(member)

    def remove_group(self, ns: str, group: str) -> None:
        self._db.pop(group, None)

    def get_members(self, ns: str, group: str) -> Set[bytes]:
        return self._db.get(group, set())

    def is_member(self, ns: str, group: str, member: bytes) -> bool:
        if group not in self._db:
            return False
        if member in self._db[group]:
            return True
        return False

    def group_size(self, ns: str, group: str) -> int:
        if group not in self._db:
            return 0
        return len(self._db[group])

    def set_and_publish(self, ns: str, channels_and_events: Dict[str, List[str]],
                        data_map: Dict[str, bytes]) -> None:
        self._db.update(data_map.copy())
        for channel, events in channels_and_events.items():
            for event in events:
                self._queue.put((channel, event))

    def set_if_and_publish(self, ns: str, channels_and_events: Dict[str, List[str]], key: str,
                           old_data: bytes, new_data: bytes) -> bool:
        if self.set_if(ns, key, old_data, new_data):
            for channel, events in channels_and_events.items():
                for event in events:
                    self._queue.put((channel, event))
            return True
        return False

    def set_if_not_exists_and_publish(self, ns: str, channels_and_events: Dict[str, List[str]],
                                      key: str, data: bytes) -> bool:
        if self.set_if_not_exists(ns, key, data):
            for channel, events in channels_and_events.items():
                for event in events:
                    self._queue.put((channel, event))
            return True
        return False

    def remove_and_publish(self, ns: str, channels_and_events: Dict[str, List[str]],
                           keys: List[str]) -> None:
        for key in keys:
            self._db.pop(key, None)
        for channel, events in channels_and_events.items():
            for event in events:
                self._queue.put((channel, event))

    def remove_if_and_publish(self, ns: str, channels_and_events: Dict[str, List[str]], key: str,
                              data: bytes) -> bool:
        if self.remove_if(ns, key, data):
            for channel, events in channels_and_events.items():
                for event in events:
                    self._queue.put((channel, event))
            return True
        return False

    def remove_all_and_publish(self, ns: str, channels_and_events: Dict[str, List[str]]) -> None:
        # Note: Since fake db has only one namespace, this deletes all keys
        self._db.clear()
        for channel, events in channels_and_events.items():
            for event in events:
                self._queue.put((channel, event))

    def subscribe_channel(self, ns: str, cb: Callable[[str, str], None],
                          channels: List[str]) -> None:
        for channel in channels:
            self._channel_cbs[channel] = cb
            if not self._listen_thread.is_alive() and self._run_in_thread:
                self._listen_thread.start()

    def _listen(self):
        while True:
            message = self._queue.get()
            cb = self._channel_cbs.get(message[0], None)
            if cb:
                cb(message[0], message[1])
            time.sleep(0.001)

    def unsubscribe_channel(self, ns: str, channels: List[str]) -> None:
        for channel in channels:
            self._channel_cbs.pop(channel, None)

    def start_event_listener(self) -> None:
        if self._listen_thread.is_alive():
            raise Exception("Event loop already started")
        if len(self._channel_cbs) > 0:
            self._listen_thread.start()
        self._run_in_thread = True

    def handle_events(self) -> Optional[Tuple[str, str]]:
        if self._listen_thread.is_alive() or self._run_in_thread:
            raise Exception("Event loop already started")
        try:
            message = self._queue.get(block=False)
        except queue.Empty:
            return None
        cb = self._channel_cbs.get(message[0], None)
        if cb:
            cb(message[0], message[1])
        return (message[0], message[1])


class FakeDictBackendLock(DbBackendLockAbc):
    """
    A class providing fake implementation of database backend lock of Shared Data Layer (SDL).
    This class does not provide working database solution, this class can be used in testing
    purposes only. Implementation does not provide shared database resource, SDL client sees
    only its local local 'fake' database, which is a simple Python dictionary. Also keys are
    stored in database under the same namespace.
    Args:
        ns (str): Namespace under which this lock is targeted.
        name (str): Lock name, identifies the lock key in a Redis database backend.
        expiration (int, float): Lock expiration time after which the lock is removed if it hasn't
                                 been released earlier by a 'release' method.
        redis_backend (FakeBackend): Database backend object containing fake databese connection.
    """

    def __init__(self, ns: str, name: str, expiration: Union[int, float],
                 redis_backend: FakeDictBackend) -> None:
        super().__init__(ns, name)
        self._locked = False
        self._ns = ns
        self._lock_name = name
        self._lock_expiration = expiration
        self.redis_backend = redis_backend

    def __str__(self):
        return str(
            {
                "lock DB type": "FAKE DB",
                "lock namespace": self._ns,
                "lock name": self._lock_name,
                "lock status": self._lock_status_to_string()
            }
        )

    def acquire(self, retry_interval: Union[int, float] = 0.1,
                retry_timeout: Union[int, float] = 10) -> bool:
        if self._locked:
            return False
        self._locked = True
        return self._locked

    def release(self) -> None:
        self._locked = False

    def refresh(self) -> None:
        pass

    def get_validity_time(self) -> Union[int, float]:
        return self._lock_expiration

    def _lock_status_to_string(self) -> str:
        if self._locked:
            return 'locked'
        return 'unlocked'



# Copyright (c) 2019 AT&T Intellectual Property.
# Copyright (c) 2018-2019 Nokia.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# This source code is part of the near-RT RIC (RAN Intelligent Controller)
# platform project (RICP).
#

"""The module provides implementation of the syncronous Shared Data Layer (SDL) interface."""
import builtins
import inspect
from typing import (Any, Callable, Dict, Set, List, Optional, Tuple, Union)
from ricsdl.configuration import _Configuration
from ricsdl.syncstorage_abc import (SyncStorageAbc, SyncLockAbc)
import ricsdl.backend
from ricsdl.backend.dbbackend_abc import DbBackendAbc
from ricsdl.exceptions import (SdlException, SdlTypeError)


def func_arg_checker(exception, start_arg_idx, **types):
    """Decorator to validate function arguments."""
    def _check(func):
        if not __debug__:
            return func

        def _validate(*args, **kwds):
            for idx, arg in enumerate(args[start_arg_idx:], start_arg_idx):
                if func.__code__.co_varnames[idx] in types and \
                        not isinstance(arg, types[func.__code__.co_varnames[idx]]):
                    raise exception(r"Wrong argument type: '{}'={}. Must be: {}".
                                    format(func.__code__.co_varnames[idx], type(arg),
                                           types[func.__code__.co_varnames[idx]]))
            for kwdname, kwdval in kwds.items():
                if kwdname in types and not isinstance(kwdval, types[kwdname]):
                    raise exception(r"Wrong argument type: '{}'={}. Must be: {}".
                                    format(kwdname, type(kwdval), types[kwdname]))
            return func(*args, **kwds)
        _validate.__name__ = func.__name__
        return _validate
    return _check


class SyncLock(SyncLockAbc):
    """
    This class implements Shared Data Layer (SDL) abstract 'SyncLockAbc' class.

    A lock instance is created per namespace and it is identified by its `name` within a namespace.

    Args:
        ns (str): Namespace under which this lock is targeted.
        name (str): Lock name, identifies the lock key in SDL storage.
        expiration (int, float): Lock expiration time after which the lock is removed if it hasn't
                                 been released earlier by a 'release' method.
        storage (SyncStorage): Database backend object containing connection to a database.
    """
    @func_arg_checker(SdlTypeError, 1, ns=str, name=str, expiration=(int, float))
    def __init__(self, ns: str, name: str, expiration: Union[int, float],
                 storage: 'SyncStorage') -> None:

        super().__init__(ns, name, expiration)
        self.__configuration = storage.get_configuration()
        self.__dbbackendlock = ricsdl.backend.get_backend_lock_instance(self.__configuration,
                                                                        ns, name, expiration,
                                                                        storage.get_backend())

    def __str__(self):
        return str(
            {
                "namespace": self._ns,
                "name": self._name,
                "expiration": self._expiration,
                "backend lock": str(self.__dbbackendlock)
            }
        )

    @func_arg_checker(SdlTypeError, 1, retry_interval=(int, float),
                      retry_timeout=(int, float))
    def acquire(self, retry_interval: Union[int, float] = 0.1,
                retry_timeout: Union[int, float] = 10) -> bool:
        return self.__dbbackendlock.acquire(retry_interval, retry_timeout)

    def release(self) -> None:
        self.__dbbackendlock.release()

    def refresh(self) -> None:
        self.__dbbackendlock.refresh()

    def get_validity_time(self) -> Union[int, float]:
        return self.__dbbackendlock.get_validity_time()


class SyncStorage(SyncStorageAbc):
    """
    This class implements Shared Data Layer (SDL) abstract 'SyncStorageAbc' class.

    This class provides synchronous access to all the namespaces in SDL storage.
    Data can be written, read and removed based on keys known to clients. Keys are unique within
    a namespace, namespace identifier is passed as a parameter to all the operations.

    Args:
        fake_db_backend (str): Optional parameter. Parameter enables fake DB backend usage for an
                               SDL instance. Fake DB backend is ONLY allowed to use for testing
                               purposes at development phase of SDL clients when more advanced
                               database services are not necessarily needed. Currently value 'dict'
                               is only allowed value for the parameter, which enables dictionary
                               type of fake DB backend.
    """
    def __init__(self, fake_db_backend=None) -> None:
        super().__init__()
        self.__configuration = _Configuration(fake_db_backend)
        self.__dbbackend = ricsdl.backend.get_backend_instance(self.__configuration)

    def __del__(self):
        self.close()

    def __str__(self):
        return str(
            {
                "configuration": str(self.__configuration),
                "backend": str(self.__dbbackend)
            }
        )

    def is_active(self):
        try:
            return self.__dbbackend.is_connected()
        except SdlException:
            return False

    def close(self):
        self.__dbbackend.close()

    @func_arg_checker(SdlTypeError, 1, ns=str, data_map=dict)
    def set(self, ns: str, data_map: Dict[str, bytes]) -> None:
        self._validate_key_value_dict(data_map)
        self.__dbbackend.set(ns, data_map)

    @func_arg_checker(SdlTypeError, 1, ns=str, key=str, old_data=bytes, new_data=bytes)
    def set_if(self, ns: str, key: str, old_data: bytes, new_data: bytes) -> bool:
        return self.__dbbackend.set_if(ns, key, old_data, new_data)

    @func_arg_checker(SdlTypeError, 1, ns=str, key=str, data=bytes)
    def set_if_not_exists(self, ns: str, key: str, data: bytes) -> bool:
        return self.__dbbackend.set_if_not_exists(ns, key, data)

    @func_arg_checker(SdlTypeError, 1, ns=str, keys=(str, builtins.set))
    def get(self, ns: str, keys: Union[str, Set[str]]) -> Dict[str, bytes]:
        disordered = self.__dbbackend.get(ns, list(keys))
        return {k: disordered[k] for k in sorted(disordered)}

    @func_arg_checker(SdlTypeError, 1, ns=str, key_pattern=str)
    def find_keys(self, ns: str, key_pattern: str) -> List[str]:
        return self.__dbbackend.find_keys(ns, key_pattern)

    @func_arg_checker(SdlTypeError, 1, ns=str, key_pattern=str)
    def find_and_get(self, ns: str, key_pattern: str) -> Dict[str, bytes]:
        disordered = self.__dbbackend.find_and_get(ns, key_pattern)
        return {k: disordered[k] for k in sorted(disordered)}

    @func_arg_checker(SdlTypeError, 1, ns=str, keys=(str, builtins.set))
    def remove(self, ns: str, keys: Union[str, Set[str]]) -> None:
        self.__dbbackend.remove(ns, list(keys))

    @func_arg_checker(SdlTypeError, 1, ns=str, key=str, data=bytes)
    def remove_if(self, ns: str, key: str, data: bytes) -> bool:
        return self.__dbbackend.remove_if(ns, key, data)

    @func_arg_checker(SdlTypeError, 1, ns=str)
    def remove_all(self, ns: str) -> None:
        keys = self.__dbbackend.find_keys(ns, '*')
        if keys:
            self.__dbbackend.remove(ns, keys)

    @func_arg_checker(SdlTypeError, 1, ns=str, group=str, members=(bytes, builtins.set))
    def add_member(self, ns: str, group: str, members: Union[bytes, Set[bytes]]) -> None:
        self.__dbbackend.add_member(ns, group, members)

    @func_arg_checker(SdlTypeError, 1, ns=str, group=str, members=(bytes, builtins.set))
    def remove_member(self, ns: str, group: str, members: Union[bytes, Set[bytes]]) -> None:
        self.__dbbackend.remove_member(ns, group, members)

    @func_arg_checker(SdlTypeError, 1, ns=str, group=str)
    def remove_group(self, ns: str, group: str) -> None:
        self.__dbbackend.remove_group(ns, group)

    @func_arg_checker(SdlTypeError, 1, ns=str, group=str)
    def get_members(self, ns: str, group: str) -> Set[bytes]:
        return self.__dbbackend.get_members(ns, group)

    @func_arg_checker(SdlTypeError, 1, ns=str, group=str, member=bytes)
    def is_member(self, ns: str, group: str, member: bytes) -> bool:
        return self.__dbbackend.is_member(ns, group, member)

    @func_arg_checker(SdlTypeError, 1, ns=str, group=str)
    def group_size(self, ns: str, group: str) -> int:
        return self.__dbbackend.group_size(ns, group)

    @func_arg_checker(SdlTypeError, 1, ns=str, channels_and_events=dict, data_map=dict)
    def set_and_publish(self, ns: str, channels_and_events: Dict[str, Union[str, List[str]]],
                        data_map: Dict[str, bytes]) -> None:
        self._validate_key_value_dict(data_map)
        self._validate_channels_events(channels_and_events)
        for channel, events in channels_and_events.items():
            channels_and_events[channel] = [events] if isinstance(events, str) else events
        self.__dbbackend.set_and_publish(ns, channels_and_events, data_map)

    @func_arg_checker(SdlTypeError,
                      1,
                      ns=str,
                      channels_and_events=dict,
                      key=str,
                      old_data=bytes,
                      new_data=bytes)
    def set_if_and_publish(self, ns: str, channels_and_events: Dict[str, Union[str, List[str]]],
                           key: str, old_data: bytes, new_data: bytes) -> bool:
        self._validate_channels_events(channels_and_events)
        for channel, events in channels_and_events.items():
            channels_and_events[channel] = [events] if isinstance(events, str) else events
        return self.__dbbackend.set_if_and_publish(ns, channels_and_events, key, old_data, new_data)

    @func_arg_checker(SdlTypeError, 1, ns=str, channels_and_events=dict, key=str, data=bytes)
    def set_if_not_exists_and_publish(self, ns: str,
                                      channels_and_events: Dict[str, Union[str, List[str]]],
                                      key: str, data: bytes) -> bool:
        self._validate_channels_events(channels_and_events)
        for channel, events in channels_and_events.items():
            channels_and_events[channel] = [events] if isinstance(events, str) else events
        return self.__dbbackend.set_if_not_exists_and_publish(ns, channels_and_events, key, data)

    @func_arg_checker(SdlTypeError, 1, ns=str, channels_and_events=dict, keys=(str, builtins.set))
    def remove_and_publish(self, ns: str, channels_and_events: Dict[str, Union[str, List[str]]],
                           keys: Union[str, Set[str]]) -> None:
        self._validate_channels_events(channels_and_events)
        for channel, events in channels_and_events.items():
            channels_and_events[channel] = [events] if isinstance(events, str) else events
        keys = [keys] if isinstance(keys, str) else list(keys)
        self.__dbbackend.remove_and_publish(ns, channels_and_events, keys)

    @func_arg_checker(SdlTypeError, 1, ns=str, channels_and_events=dict, key=str, data=bytes)
    def remove_if_and_publish(self, ns: str, channels_and_events: Dict[str, Union[str, List[str]]],
                              key: str, data: bytes) -> bool:
        self._validate_channels_events(channels_and_events)
        for channel, events in channels_and_events.items():
            channels_and_events[channel] = [events] if isinstance(events, str) else events
        return self.__dbbackend.remove_if_and_publish(ns, channels_and_events, key, data)

    @func_arg_checker(SdlTypeError, 1, ns=str, channels_and_events=dict)
    def remove_all_and_publish(self, ns: str,
                               channels_and_events: Dict[str, Union[str, List[str]]]) -> None:
        self._validate_channels_events(channels_and_events)
        for channel, events in channels_and_events.items():
            channels_and_events[channel] = [events] if isinstance(events, str) else events
        self.__dbbackend.remove_all_and_publish(ns, channels_and_events)

    @func_arg_checker(SdlTypeError, 1, ns=str, cb=Callable, channels=(str, builtins.set))
    def subscribe_channel(self, ns: str, cb: Callable[[str, str], None],
                          channels: Union[str, Set[str]]) -> None:
        self._validate_callback(cb)
        channels = [channels] if isinstance(channels, str) else list(channels)
        self.__dbbackend.subscribe_channel(ns, cb, channels)

    @func_arg_checker(SdlTypeError, 1, ns=str, channels=(str, builtins.set))
    def unsubscribe_channel(self, ns: str, channels: Union[str, Set[str]]) -> None:
        channels = [channels] if isinstance(channels, str) else list(channels)
        self.__dbbackend.unsubscribe_channel(ns, channels)

    def start_event_listener(self) -> None:
        self.__dbbackend.start_event_listener()

    def handle_events(self) -> Optional[Tuple[str, str]]:
        return self.__dbbackend.handle_events()

    @func_arg_checker(SdlTypeError, 1, ns=str, resource=str, expiration=(int, float))
    def get_lock_resource(self, ns: str, resource: str, expiration: Union[int, float]) -> SyncLock:
        return SyncLock(ns, resource, expiration, self)

    def get_backend(self) -> DbBackendAbc:
        """Return backend instance."""
        return self.__dbbackend

    def get_configuration(self) -> _Configuration:
        """Return configuration what was valid when the SDL instance was initiated."""
        return self.__configuration

    @classmethod
    def _validate_key_value_dict(cls, kv):
        for k, v in kv.items():
            if not isinstance(k, str):
                raise SdlTypeError(r"Wrong dict key type: {}={}. Must be: str".format(k, type(k)))
            if not isinstance(v, bytes):
                raise SdlTypeError(r"Wrong dict value type: {}={}. Must be: bytes".format(v, type(v)))

    @classmethod
    def _validate_channels_events(cls, channels_and_events: Dict[Any, Any]):
        for channel, events in channels_and_events.items():
            if not isinstance(channel, str):
                raise SdlTypeError(r"Wrong channel type: {}={}. Must be: str".format(
                    channel, type(channel)))
            if not isinstance(events, (list, str)):
                raise SdlTypeError(r"Wrong event type: {}={}. Must be: str".format(
                    events, type(events)))
            if isinstance(events, list):
                for event in events:
                    if not isinstance(event, str):
                        raise SdlTypeError(r"Wrong event type: {}={}. Must be: str".format(
                            events, type(events)))

    @classmethod
    def _validate_callback(cls, cb):
        param_len = len(inspect.signature(cb).parameters)
        if param_len != 2:
            raise SdlTypeError(
                f"Callback function should take 2 positional argument but {param_len} were given")



# Copyright (c) 2019 AT&T Intellectual Property.
# Copyright (c) 2018-2019 Nokia.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# This source code is part of the near-RT RIC (RAN Intelligent Controller)
# platform project (RICP).
#


"""The module provides Shared Data Layer (SDL) database backend interface."""

from typing import (Callable, Dict, Set, List, Optional, Tuple, Union)
from abc import ABC, abstractmethod


class DbBackendAbc(ABC):
    """An abstract Shared Data Layer (SDL) class providing database backend interface."""

    @abstractmethod
    def is_connected(self):
        """Test database backend connection."""
        pass

    @abstractmethod
    def close(self):
        """Close database backend connection."""
        pass

    @abstractmethod
    def set(self, ns: str, data_map: Dict[str, bytes]) -> None:
        """Write key value data mapping to database under a namespace."""
        pass

    @abstractmethod
    def set_if(self, ns: str, key: str, old_data: bytes, new_data: bytes) -> bool:
        """"Write key value to database under a namespace if the old value is expected one."""
        pass

    @abstractmethod
    def set_if_not_exists(self, ns: str, key: str, data: bytes) -> bool:
        """"Write key value to database under a namespace if key doesn't exist."""
        pass

    @abstractmethod
    def get(self, ns: str, keys: List[str]) -> Dict[str, bytes]:
        """"Return values of the keys under a namespace."""
        pass

    @abstractmethod
    def find_keys(self, ns: str, key_pattern: str) -> List[str]:
        """"Return all the keys matching search pattern under a namespace in database."""
        pass

    @abstractmethod
    def find_and_get(self, ns: str, key_pattern: str) -> Dict[str, bytes]:
        """
        Return all the keys with their values matching search pattern under a namespace in
        database.
        """
        pass

    @abstractmethod
    def remove(self, ns: str, keys: List[str]) -> None:
        """Remove keys and their data from database."""
        pass

    @abstractmethod
    def remove_if(self, ns: str, key: str, data: bytes) -> bool:
        """
           Remove key and its data from database if if the current data value is expected
           one.
        """
        pass

    @abstractmethod
    def add_member(self, ns: str, group: str, members: Set[bytes]) -> None:
        """Add new members to a group under a namespace in database."""
        pass

    @abstractmethod
    def remove_member(self, ns: str, group: str, members: Set[bytes]) -> None:
        """Remove members from a group under a namespace in database."""
        pass

    @abstractmethod
    def remove_group(self, ns: str, group: str) -> None:
        """Remove a group under a namespace in database along with it's members."""
        pass

    @abstractmethod
    def get_members(self, ns: str, group: str) -> Set[bytes]:
        """Get all the members of a group under a namespace in database."""
        pass

    @abstractmethod
    def is_member(self, ns: str, group: str, member: bytes) -> bool:
        """Validate if a given member is in the group under a namespace in database."""
        pass

    @abstractmethod
    def group_size(self, ns: str, group: str) -> int:
        """Return the number of members in a group under a namespace in database."""
        pass

    @abstractmethod
    def set_and_publish(self, ns: str, channels_and_events: Dict[str, List[str]],
                        data_map: Dict[str, bytes]) -> None:
        """Publish event to channel after writing data."""
        pass

    @abstractmethod
    def set_if_and_publish(self, ns: str, channels_and_events: Dict[str, List[str]], key: str,
                           old_data: bytes, new_data: bytes) -> bool:
        """
        Publish event to channel after writing key value to database under a namespace
        if the old value is expected one.
        """
        pass

    @abstractmethod
    def set_if_not_exists_and_publish(self, ns: str, channels_and_events: Dict[str, List[str]],
                                      key: str, data: bytes) -> bool:
        """"
        Publish event to channel after writing key value to database under a namespace if
        key doesn't exist.
        """
        pass

    @abstractmethod
    def remove_and_publish(self, ns: str, channels_and_events: Dict[str, List[str]],
                           keys: List[str]) -> None:
        """Publish event to channel after removing data."""
        pass

    @abstractmethod
    def remove_if_and_publish(self, ns: str, channels_and_events: Dict[str, List[str]], key: str,
                              data: bytes) -> bool:
        """
        Publish event to channel after removing key and its data from database if the
        current data value is expected one.
        """
        pass

    @abstractmethod
    def remove_all_and_publish(self, ns: str, channels_and_events: Dict[str, List[str]]) -> None:
        """
        Publish event to channel after removing all keys in namespace.
        """
        pass

    @abstractmethod
    def subscribe_channel(self, ns: str, cb: Callable[[str, str], None],
                          channels: List[str]) -> None:
        """
        This takes a callback function and one or many channels to be subscribed.
        When an event is received for the given channel, the given callback function
        shall be called with channel and notifications as parameter.
        """
        pass

    @abstractmethod
    def unsubscribe_channel(self, ns: str, channels: List[str]) -> None:
        """Unsubscribes from channel and removes set callback function."""
        pass

    @abstractmethod
    def start_event_listener(self) -> None:
        """
        start_event_listener creates an event loop in a separate thread for handling
        notifications from subscriptions.
        """
        pass

    @abstractmethod
    def handle_events(self) -> Optional[Tuple[str, str]]:
        """
        handle_events is a non-blocking function that returns a tuple containing channel
        name and message received from notification.
        """
        pass


class DbBackendLockAbc(ABC):
    """
    An abstract Shared Data Layer (SDL) class providing database backend lock interface.
    Args:
        ns (str): Namespace under which this lock is targeted.
        name (str): Lock name, identifies the lock key in a database backend.
    """
    def __init__(self, ns: str, name: str) -> None:
        self._ns = ns
        self._lock_name = name
        super().__init__()

    @abstractmethod
    def acquire(self, retry_interval: Union[int, float] = 0.1,
                retry_timeout: Union[int, float] = 10) -> bool:
        """Acquire a database lock."""
        pass

    @abstractmethod
    def release(self) -> None:
        """Release a database lock."""
        pass

    @abstractmethod
    def refresh(self) -> None:
        """Refresh the remaining validity time of the database lock back to a initial value."""
        pass

    @abstractmethod
    def get_validity_time(self) -> Union[int, float]:
        """Return remaining validity time of the lock in seconds."""
        pass



# Copyright (c) 2020 Samsung Electronics
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# This source code is part of the near-RT RIC (RAN Intelligent Controller)
# platform project (RICP).
#

"""Examples on how to use Shared Data Layer (SDL) notification feature.

Execution of  these examples requires:
 * Following Redis extension commands have been installed to runtime environment:
   - MSETPUB
   - SETIE
   - SETIEMPUB
   - SETNXMPUB
   - DELMPUB
   - DELIE
   - DELIEMPUB
   Redis v4.0 or greater is required. Older versions do not support extension modules.
   Implementation of above commands is produced by RIC DBaaS:
   https://gerrit.o-ran-sc.org/r/admin/repos/ric-plt/dbaas
   In official RIC deployments these commands are installed by `dbaas` service to Redis
   container(s).
   In development environment you may want install commands manually to pod/container, which is
   running Redis.
 * Following environment variables are needed to set to the pod/container where the application
   utilizing SDL is going to be run.
     DBAAS_SERVICE_HOST = [redis server address]
     DBAAS_SERVICE_PORT= [redis server port]
     DBAAS_MASTER_NAME = [master Redis sentinel name]. Needed to set only if sentinel is in use.
     DBAAS_SERVICE_SENTINEL_PORT = [Redis sentinel port number]. Needed to set only if sentinel
     is in use.
"""

import threading
import time

from ricsdl.syncstorage import SyncStorage
from ricsdl.exceptions import RejectedByBackend, NotConnected, BackendError

# There are two available methods for applications to handle notifications:
#   - EVENT_LISTENER (true):
#     - User calls sdl.start_event_listener() which will create an SDL managed
#       event loop for handling messages.
#   - EVENT_LISTENER (false):
#     - User need to call sdl.handle_messages() which will return the message
#
# Note: In both cases, the given callback function will be executed.
EVENT_LISTENER = True

# Constants used in the examples below.
MY_NS = 'my_ns'
MY_CHANNEL = "my_channel"
MY_LOCK = threading.Lock()


def _try_func_return(func):
    """
    Generic wrapper function to call SDL API function and handle exceptions if they are raised.
    """
    try:
        return func()
    except RejectedByBackend as exp:
        print(f'SDL function {func.__name__} failed: {str(exp)}')
        # Permanent failure, just forward the exception
        raise
    except (NotConnected, BackendError) as exp:
        print(f'SDL function {func.__name__} failed for a temporal error: {str(exp)}')
        # Here we could have a retry logic


def _try_func_callback_return(func):
    """Generic wrapper function for testing SDL APIs with callback functions.

    threading.Lock is unlocked in the callback function and threading.Lock is
    only used to demonstrate that the callback function was called.
    """
    global MY_LOCK
    MY_LOCK.acquire()
    ret = _try_func_return(func)
    while MY_LOCK.locked():
        time.sleep(0.01)
    return ret


# Creates SDL instance. The call creates connection to the SDL database backend.
mysdl = _try_func_return(SyncStorage)

# Stores the last received channel and message
last_cb_channel = ""
last_cb_message = ""

# Allows program to stop receive thread at the end of execution
stop_thread = False


def cb(channel: str, message: str):
    """An example of function that will be called when an event is received.

    This function sets last_cb_channel and last_cb_message as channel and
    message respectively. This also unlocks the global lock variable.

    Args:
        channel: Channel where the message was received
        message: Received message
    """
    global last_cb_channel, last_cb_message, MY_LOCK
    last_cb_channel = channel
    last_cb_message = message
    if MY_LOCK.locked():
        MY_LOCK.release()


def listen_thread():
    """An example of a listener thread that continuously calls sdl.handle_events()."""
    global mysdl
    global stop_thread
    while not stop_thread:
        message = mysdl.handle_events()
        if message:
            # You could process message here
            pass
        time.sleep(0.001)

# Subscribe to MY_CHANNEL. We expect that anytime we receive a message in the
# channel, cb function will be called.
_try_func_return(lambda: mysdl.subscribe_channel(MY_NS, cb, MY_CHANNEL))

# As mentioned above, there are two available methods for applications to
# handle notifications
if EVENT_LISTENER:
    _try_func_return(mysdl.start_event_listener)
else:
    thread = threading.Thread(target=listen_thread)
    thread.start()

# Sets a value 'my_value' for a key 'my_key' under given namespace. Note that value
# type must be bytes and multiple key values can be set in one set function call.
_try_func_callback_return(
    lambda: mysdl.set_and_publish(MY_NS, {MY_CHANNEL: "SET PUBLISH"}, {'my_key': b'my_value'}))
assert last_cb_channel == MY_CHANNEL and last_cb_message == "SET PUBLISH"

# Sets a value 'my_value2' for a key 'my_key' under given namespace only if the old value is
# 'my_value'.
was_set = _try_func_callback_return(lambda: mysdl.set_if_and_publish(
    MY_NS, {MY_CHANNEL: "SET IF PUBLISH"}, 'my_key', b'my_value', b'my_value2'))
assert was_set is True
assert last_cb_channel == MY_CHANNEL and last_cb_message == "SET IF PUBLISH"
# Try again. This time value 'my_value2' won't be set, because the key has already 'my_value2'
# value. Callback function will not be called here.
was_set = _try_func_return(lambda: mysdl.set_if_and_publish(MY_NS, {MY_CHANNEL: "SET IF PUBLISH"},
                                                            'my_key', b'my_value', b'my_value2'))
assert was_set is False

# Sets a value 'my_value' for a key 'my_key2' under given namespace only if the key doesn't exist.
# Note that value types must be bytes.
was_set = _try_func_callback_return(lambda: mysdl.set_if_not_exists_and_publish(
    MY_NS, {MY_CHANNEL: "SET IF NOT PUBLISH"}, 'my_key2', b'my_value'))
assert was_set is True
assert last_cb_channel == MY_CHANNEL and last_cb_message == "SET IF NOT PUBLISH"
# Try again. This time the key 'my_key2' already exists. Callback function will not be called here.
was_set = _try_func_return(lambda: mysdl.set_if_not_exists_and_publish(
    MY_NS, {MY_CHANNEL: "SET IF NOT PUBLISH"}, 'my_key2', b'my_value'))
assert was_set is False

# Removes a key 'my_key' under given namespace.
_try_func_callback_return(
    lambda: mysdl.remove_and_publish(MY_NS, {MY_CHANNEL: "REMOVE PUBLISH"}, 'my_key'))
my_ret_dict = _try_func_return(lambda: mysdl.get(MY_NS, 'my_key'))
assert my_ret_dict == {}
assert last_cb_channel == MY_CHANNEL and last_cb_message == "REMOVE PUBLISH"

# Removes a key 'my_key' under given namespace only if the old value is 'my_value'.
was_removed = _try_func_callback_return(lambda: mysdl.remove_if_and_publish(
    MY_NS, {MY_CHANNEL: "REMOVE IF PUBLISH"}, 'my_key2', b'my_value'))
assert was_removed is True
assert last_cb_channel == MY_CHANNEL and last_cb_message == "REMOVE IF PUBLISH"
# Try again to remove not anymore existing key 'my_key'. Callback function will not be called here.
was_removed = _try_func_return(lambda: mysdl.remove_if_and_publish(
    MY_NS, {MY_CHANNEL: "REMOVE IF PUBLISH"}, 'my_key2', b'my_value'))
assert was_removed is False

# Removes all the keys under given namespace.
_try_func_return(lambda: mysdl.set(MY_NS, {'my_key': b'something'}))
my_ret_dict = _try_func_return(lambda: mysdl.get(MY_NS, {'my_key'}))
assert my_ret_dict != {}

_try_func_callback_return(
    lambda: mysdl.remove_all_and_publish(MY_NS, {MY_CHANNEL: "REMOVE ALL PUBLISH"}))
my_ret_dict = _try_func_return(lambda: mysdl.get(MY_NS, {'my_key'}))
assert my_ret_dict == {}
assert last_cb_channel == MY_CHANNEL and last_cb_message == "REMOVE ALL PUBLISH"

stop_thread = True
mysdl.close()



# Copyright (c) 2019 AT&T Intellectual Property.
# Copyright (c) 2018-2019 Nokia.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# This source code is part of the near-RT RIC (RAN Intelligent Controller)
# platform project (RICP).
#


"""
Examples how to use synchronous API functions of the Shared Data Layer (SDL).
Execution of  these examples requires:
 * Following Redis extension commands have been installed to runtime environment:
   - MSETPUB
   - SETIE
   - SETIEMPUB
   - SETNXMPUB
   - DELMPUB
   - DELIE
   - DELIEMPUB
   Redis v4.0 or greater is required. Older versions do not support extension modules.
   Implementation of above commands is produced by RIC DBaaS:
   https://gerrit.o-ran-sc.org/r/admin/repos/ric-plt/dbaas
   In official RIC deployments these commands are installed by `dbaas` service to Redis
   container(s).
   In development environment you may want install commands manually to pod/container, which is
   running Redis.
 * Following environment variables are needed to set to the pod/container where the application
   utilizing SDL is going to be run.
     DBAAS_SERVICE_HOST = [redis server address]
     DBAAS_SERVICE_PORT= [redis server port]
     DBAAS_MASTER_NAME = [master Redis sentinel name]. Needed to set only if sentinel is in use.
     DBAAS_SERVICE_SENTINEL_PORT = [Redis sentinel port number]. Needed to set only if sentinel
     is in use.
"""
from ricsdl.syncstorage import SyncStorage
from ricsdl.exceptions import RejectedByBackend, NotConnected, BackendError


# Constants used in the examples below.
MY_NS = 'my_ns'
MY_GRP_NS = 'my_group_ns'
MY_LOCK_NS = 'my_group_ns'


def _try_func_return(func):
    """
    Generic wrapper function to call SDL API function and handle exceptions if they are raised.
    """
    try:
        return func()
    except RejectedByBackend as exp:
        print(f'SDL function {func.__name__} failed: {str(exp)}')
        # Permanent failure, just forward the exception
        raise
    except (NotConnected, BackendError) as exp:
        print(f'SDL function {func.__name__} failed for a temporal error: {str(exp)}')
        # Here we could have a retry logic


# Creates SDL instance. The call creates connection to the SDL database backend.
mysdl = _try_func_return(SyncStorage)

# Creates SDL instance what utilizes a fake database backend. Fake database is meant to
# be used only at development phase of SDL clients. It does not provide more advanced
# database services.
# mysdl = _try_func_return(lambda: SyncStorage(fake_db_backend='dict'))

# Checks if SDL is operational. Note that it is not necessary to call `is_active()` after each
# SDL instance creation. Below example is here just to show how to call it spontaneously
# when SDL healthiness is needed to check.
is_active = mysdl.is_active()
assert is_active is True

# Sets a value 'my_value' for a key 'my_key' under given namespace. Note that value
# type must be bytes and multiple key values can be set in one set function call.
_try_func_return(lambda: mysdl.set(MY_NS, {'my_key': b'my_value'}))


# Gets the value of 'my_value' under given namespace.
# Note that the type of returned value is bytes.
my_ret_dict = _try_func_return(lambda: mysdl.get(MY_NS, {'my_key', 'someting not existing'}))
for key, val in my_ret_dict.items():
    assert val.decode("utf-8") == u'my_value'


# Sets a value 'my_value2' for a key 'my_key' under given namespace only if the old value is
# 'my_value'.
# Note that value types must be bytes.
was_set = _try_func_return(lambda: mysdl.set_if(MY_NS, 'my_key', b'my_value', b'my_value2'))
assert was_set is True
# Try again. This time value 'my_value2' won't be set, because the key has already 'my_value2'
# value.
was_set = _try_func_return(lambda: mysdl.set_if(MY_NS, 'my_key', b'my_value', b'my_value2'))
assert was_set is False


# Sets a value 'my_value' for a key 'my_key2' under given namespace only if the key doesn't exist.
# Note that value types must be bytes.
was_set = _try_func_return(lambda: mysdl.set_if_not_exists(MY_NS, 'my_key2', b'my_value'))
assert was_set is True
# Try again. This time the key 'my_key2' already exists.
was_set = _try_func_return(lambda: mysdl.set_if_not_exists(MY_NS, 'my_key2', b'my_value'))
assert was_set is False


# Removes a key 'my_key' under given namespace.
_try_func_return(lambda: mysdl.remove(MY_NS, 'my_key'))
my_ret_dict = _try_func_return(lambda: mysdl.get(MY_NS, 'my_key'))
assert my_ret_dict == {}


# Removes a key 'my_key' under given namespace only if the old value is 'my_value'.
was_removed = _try_func_return(lambda: mysdl.remove_if(MY_NS, 'my_key2', b'my_value'))
assert was_removed is True
# Try again to remove not anymore existing key 'my_key'.
was_removed = _try_func_return(lambda: mysdl.remove_if(MY_NS, 'my_key2', b'my_value'))
assert was_removed is False


# Removes all the keys under given namespace.
_try_func_return(lambda: mysdl.set(MY_NS, {'my_key': b'something'}))
my_ret_dict = _try_func_return(lambda: mysdl.get(MY_NS, {'my_key'}))
assert my_ret_dict != {}

_try_func_return(lambda: mysdl.remove_all(MY_NS))
my_ret_dict = _try_func_return(lambda: mysdl.get(MY_NS, {'my_key'}))
assert my_ret_dict == {}


# Finds keys under given namespace that are matching to given key prefix 'my_k'.
_try_func_return(lambda: mysdl.set(MY_NS, {'my_key': b'my_value'}))
ret_keys = _try_func_return(lambda: mysdl.find_keys(MY_NS, 'my_k*'))
assert ret_keys == ['my_key']


# Finds keys and their values under given namespace that are matching to given key search
# pattern 'my_k*'.
# Note that the type of returned value is bytes.
ret_key_values = _try_func_return(lambda: mysdl.find_and_get(MY_NS, 'my_k*'))
assert ret_key_values == {'my_key': b'my_value'}

_try_func_return(lambda: mysdl.remove_all(MY_NS))


# Adds a member 'a' to a group 'my_group' under given namespace. A group is a unique collection of
# members.
# Note that member type must be bytes and multiple members can be set in one set function call.
_try_func_return(lambda: mysdl.add_member(MY_GRP_NS, 'my_group', {b'a'}))
# Try again to add a member 'a'. This time 'a' won't be added, because 'a' belongs already to
# the group.
_try_func_return(lambda: mysdl.add_member(MY_GRP_NS, 'my_group', {b'a'}))


# Gets group 'my_group' members under given namespace.
# Note that the type of returned member is bytes.
ret_members = _try_func_return(lambda: mysdl.get_members(MY_GRP_NS, 'my_group'))
assert ret_members == {b'a'}


# Checks if 'a' is a member of the group 'my_group' under given namespace.
was_member = _try_func_return(lambda: mysdl.is_member(MY_GRP_NS, 'my_group', b'a'))
assert was_member is True
was_member = _try_func_return(lambda: mysdl.is_member(MY_GRP_NS, 'my_group', b'not a member'))
assert was_member is False


# Returns the count of members of a group 'my_group' under given namespace.
ret_count = _try_func_return(lambda: mysdl.group_size(MY_GRP_NS, 'my_group'))
assert ret_count == 1


# Removes the member 'a' of the group 'my_group' under given namespace.
_try_func_return(lambda: mysdl.remove_member(MY_GRP_NS, 'my_group', {b'a', b'not exists'}))
ret_count = _try_func_return(lambda: mysdl.group_size(MY_GRP_NS, 'my_group'))
assert ret_count == 0


# Removes the group 'my_group' under given namespace.
_try_func_return(lambda: mysdl.add_member(MY_GRP_NS, 'my_group', {b'a', b'b', b'c'}))
ret_count = _try_func_return(lambda: mysdl.group_size(MY_GRP_NS, 'my_group'))
assert ret_count == 3

_try_func_return(lambda: mysdl.remove_group(MY_GRP_NS, 'my_group'))
ret_count = _try_func_return(lambda: mysdl.group_size(MY_GRP_NS, 'my_group'))
ret_members = _try_func_return(lambda: mysdl.get_members(MY_GRP_NS, 'my_group'))
assert ret_count == 0
assert ret_members == set()


# Gets a lock 'my_lock' resource under given namespace.
# Note that this function does not take a lock, you need to call 'acquire' function to take
# the lock to yourself.
my_lock = _try_func_return(lambda: mysdl.get_lock_resource(MY_LOCK_NS, "my_lock", expiration=5.5))
assert my_lock is not None


# Acquires a lock from the lock resource. Return True if lock was taken within given retry limits.
was_acquired = _try_func_return(lambda: my_lock.acquire(retry_interval=0.5, retry_timeout=2))
assert was_acquired is True
# Try again. This time a lock won't be acquired successfully, because we have a lock already.
was_acquired = _try_func_return(lambda: my_lock.acquire(retry_interval=0.1, retry_timeout=0.2))
assert was_acquired is False


# Refreshs the remaining validity time of the existing lock back to the initial value.
_try_func_return(my_lock.refresh)


# Gets the remaining validity time of the lock.
ret_time = _try_func_return(my_lock.get_validity_time)
assert ret_time != 0


# Releases the lock.
_try_func_return(my_lock.release)


# Locking example what utilizes python 'with' statement with SDL lock.
# The lock is released automatically when we are out of the scope of
# 'the with my_lock' statement.
my_lock = _try_func_return(lambda: mysdl.get_lock_resource(MY_LOCK_NS, "my_lock", 2.5))
with my_lock:
    # Just an example how to use lock API
    time_left = _try_func_return(my_lock.get_validity_time)

    # Add here operations what needs to be done under a lock, for example some
    # operations with a shared resources what needs to be done in a mutually
    # exclusive way.

# Lock is not anymore hold here


# Closes the SDL connection.
mysdl.close()



# Copyright (c) 2019 AT&T Intellectual Property.
# Copyright (c) 2018-2019 Nokia.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# This source code is part of the near-RT RIC (RAN Intelligent Controller)
# platform project (RICP).
#


"Exceptions raised by the Shared Data Layer (SDL)."


class SdlTypeError(TypeError):
    """
    Exception for passing a function argument of wrong type.
    It is likely that the same request will fail repeatedly. It is advised to investigate the exact
    reason for the failure from the logs.
    """
    pass


class SdlException(Exception):
    """Base exception class for Shared Data Layer (SDL) exceptions."""
    pass


class NotConnected(SdlException):
    """
    Exception for SDL not being connected to the database backend.
    SDL is not connected to the backend data storage and therefore could not deliver the request
    to the backend data storage. Data in the backend data storage has not been altered.
    Client is advised to try the operation again later.
    """
    pass


class BackendError(SdlException):
    """
    Exception for request processing failure in SDL database backend.
    In case of a write type request, data in the backend data storage may or may not have been
    altered. Client is advised to try the operation again later.
    """
    pass


class RejectedByBackend(SdlException):
    """
    Exception for SDL database backend rejecting the request.
    Backend data storage rejected the request. In case of a write type request, data in the backend
    data storage may or may not have been altered. It is likely that the same request will fail
    repeatedly. It is advised to investigate the exact reason for the failure from the logs.
    """
    pass



# Copyright (c) 2019 AT&T Intellectual Property.
# Copyright (c) 2018-2019 Nokia.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# This source code is part of the near-RT RIC (RAN Intelligent Controller)
# platform project (RICP).
#

"""Setup script of Shared Data Layer (SDL) package."""

from os.path import dirname, abspath, join as path_join
from setuptools import setup, find_packages
from ricsdl import __version__

SETUP_DIR = abspath(dirname(__file__))


def _long_descr():
    """Yields the content of documentation files for the long description"""
    try:
        doc_path = path_join(SETUP_DIR, "README.md")
        with open(doc_path) as file:
            return file.read()
    except FileNotFoundError:  # this happens during unit testing, we don't need it
        return ""


setup(
    name="ricsdl",
    version=__version__,
    packages=find_packages(exclude=["tests.*", "tests"]),
    author="Timo Tietavainen",
    author_email='timo.tietavainen@nokia.com',
    license='Apache 2.0',
    description="Shared Data Layer (SDL) provides a high-speed interface to access shared storage",
    url="https://gerrit.o-ran-sc.org/r/admin/repos/ric-plt/sdlpy",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Telecommunications Industry",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires=">=3.7",
    keywords="RIC SDL",
    install_requires=[
        'setuptools',
        'redis',
        'hiredis'
    ],
    long_description=_long_descr(),
    long_description_content_type="text/markdown",
)
